#include <iostream>
using std::cout;
using std::endl;
using std::ostream;

#include <gmp.h>
#include <gmpxx.h>

namespace prime_gen {
	static const int attempts_for_miller_rabin = 25;

	int generate_random_prime( mpz_t p, const mp_bitcnt_t n, gmp_randstate_t state, int attempts = 800 ) {
		//this is the rejection sampling method for generating primes as discussed in the lecture
		while( attempts > 0 ) {
			--attempts;
			mpz_urandomb(p,state,n); //this gives n "pseudorandom" bits [0..n-1]
			mpz_setbit(p,n); //set the n-th bit to 1 to make it a (n+1)-bit number
			mpz_setbit(p,0); //set the 0-th bit to 0 to make it an odd number
			
			//run a variant of the miller-rabin test to check if p passes as a prime
			//is_prime == 0 means 'composite'
			//is_prime == 1 means 'probably prime'
			//is_prime == 2 means 'certainly a prime'
			const int is_prime = mpz_probab_prime_p( p, attempts_for_miller_rabin );
			
			if(  is_prime > 0 )
				return is_prime;
		}
		return 0;
	}

	int generate_random_safe_prime( mpz_t p, mpz_t q, const mp_bitcnt_t n, gmp_randstate_t state, int attempts = 800 ) {
		//this uses rejection sampling to first produce a prime q (or some composite that passes MR)
		//than it sets p = 2q+1 and runs MR again on p to check if it passes as a (safe) prime
		const int attempts_for_prime = attempts;
		
		while( attempts > 0 ) {
			--attempts;
			
			const int q_is_prime = generate_random_prime(q,n,state,attempts_for_prime);
			
			if( q_is_prime == 0 )
				continue;
				
			mpz_mul_2exp(p,q,1);
			mpz_setbit(p,0);
			
			const int p_is_prime = mpz_probab_prime_p( p, attempts_for_miller_rabin );
			
			if(  p_is_prime > 0 )
				return (p_is_prime << 2 | q_is_prime);
		}
		return 0;
	}
}

namespace dlp {
	struct GroupDescription { //only for Z_p^* and quadratic residues
		mpz_class p; //modulus, a prime
		mpz_class g; //generator
		mpz_class o; //order of the group
		
		void set( const GroupDescription& gd ) {
			p = gd.p; g = gd.g; o = gd.o;
		}
	};

	struct DHPublicKey : public GroupDescription {
		mpz_class y; //the public information in the DH protocol, i.e. y = g^x mod p with x the secret exponent
		
		void print() const {
			cout << "p = " << p << endl;
			cout << "g = " << g << endl;
			cout << "o = " << o << endl;
			cout << "y = " << y << endl;
		}
	};

	struct DHSecretKey : public GroupDescription {
		mpz_class x; //the secret exponent in the DH protocol, x in Z_o
		
		void print() const {
			cout << "p = " << p << endl;
			cout << "g = " << g << endl;
			cout << "o = " << o << endl;
			cout << "x = " << x << endl;
		}
	};
	

	void generate_quadratic_residue_description(mp_bitcnt_t n, GroupDescription& gd, gmp_randstate_t state) {
		//this generates the group description for a quadratic residue group of size q >= 2^n
		//it simply tries to find a safe prime p = 2 q +1 using rejection sampliny (see above)
		//recall: modulo a safe prime p, the quadratic residues QRp modulo p have size q with q a prime
		//so any quadratic residue except for the 1 is a generator of QRp
		//in particular 4 is always a generator of QRp for p a safe prime
		prime_gen::generate_random_safe_prime(gd.p.get_mpz_t(),gd.o.get_mpz_t(),n,state); 
		gd.g = 4;
	}
	
	void sample(mpz_class& x, const GroupDescription& gd, gmp_randstate_t state) {
		//this simply chooses an exponent x uniformaly at 'random' from the possible exponents Z_o
		mpz_urandomm(x.get_mpz_t(),state,gd.o.get_mpz_t());
	}
	
	void compute_image(mpz_class& y, const mpz_class& x, const GroupDescription& gd) {
		//this simply computes y = g^x mod p
		mpz_powm(y.get_mpz_t(),gd.g.get_mpz_t(),x.get_mpz_t(),gd.p.get_mpz_t());
	}

	void generate_keys(const GroupDescription& gd, DHSecretKey& sk, DHPublicKey& pk, gmp_randstate_t state) {
		//computation of Bob's secret and public keys
		
		//copy the group description for convenience into the keys
		pk.set(gd);
		sk.set(gd);
		
		//create the secret exponent sk.x
		sample(sk.x,gd,state);
		
		//hide the secret in the exponent to obtain the public group element pk.y
		compute_image(pk.y,sk.x,gd);
	}
	
	void compute_shared_secret_receiver(mpz_class& k, const DHSecretKey& sk, const mpz_class& pub) {
		//compute the shared secret for Bob using the public group element he received from Alice:
		//k = pub^sk.x mod sk.p
		mpz_powm(k.get_mpz_t(),pub.get_mpz_t(),sk.x.get_mpz_t(),sk.p.get_mpz_t());
	}
	
	void compute_shared_secret_sender(mpz_class& k, const DHPublicKey& pk, const mpz_class& sec) {
		//compute the shared secret for Alice using Bob's public key and the secret generated by Alice:
		//k = pk.y^sec mod pk.p
		mpz_powm(k.get_mpz_t(),pk.y.get_mpz_t(),sec.get_mpz_t(),pk.p.get_mpz_t());
	}
}

struct ElgamalCiphertext {
	mpz_class kem_data;
	mpz_class dem_data;

	void enc( const mpz_class& plaintext, const dlp::DHPublicKey& pk, gmp_randstate_t state) {
		//Elgmal encryption
		
		//first generate the secret exponent for Alice
		mpz_class secret;
		dlp::sample(secret,pk,state);
	
		//hide the secret in the exponent of the group to obtain the group element 
		//which is sent over the public channel to Bob
		dlp::compute_image(kem_data,secret,pk);
	
		//derive the shared secret
		mpz_class key;
		dlp::compute_shared_secret_sender(key,pk,secret);
		
		//use the shared secret as a one-time pad to permute the actual message element
		dem_data = key * plaintext;
		dem_data %= pk.p;
	}
	
	void dec( mpz_class& plaintext, const dlp::DHSecretKey& sk ) const {		
        //TODO
	}
	
	void print() const {
		cout << "kem data = " << kem_data << endl;
		cout << "dem_data = " << dem_data << endl;
	}
};

void dh_example(gmp_randstate_t state) {
	//just an example
	
	//2^n is the order of the quadratic residues
	const mp_bitcnt_t n = 256;
	
	
	dlp::GroupDescription 	gd;
	dlp::DHPublicKey		pk;
	dlp::DHSecretKey		sk;
	
	//generate the description of a quadratic residue group of order >= 2^n
	dlp::generate_quadratic_residue_description(n,gd,state);
	
	//compute Bob's public and secret key
	dlp::generate_keys(gd,sk,pk,state);

	cout << "Bob's secret key:" << endl;
	sk.print();
	
	cout << "Bob's public key:" << endl;
	pk.print();
	
	//generate Alice's secret exponent
	mpz_class alices_secret_exp;
	dlp::sample(alices_secret_exp,pk,state);
	
	cout << "Alice's secret exponent:\n" << alices_secret_exp << endl;
	
	//comput the shared secret on Alice' side
	mpz_class alices_shared_secret;
	dlp::compute_shared_secret_sender(alices_shared_secret,pk,alices_secret_exp);
	
	cout << "Alice's shared secret:\n" << alices_shared_secret << endl;
	
	//compute the element that Alice sends to Bob over the public channel
	mpz_class alices_public_elem;
	dlp::compute_image(alices_public_elem,alices_secret_exp,pk);
	
	cout << "Alice's public group element:\n" << alices_public_elem << endl;
	
	//compute the shared secret on Bob's side
	mpz_class bobs_shared_secret;
	dlp::compute_shared_secret_receiver(bobs_shared_secret,sk,alices_public_elem);
	
	cout << "Bob's shared secret:\n" << bobs_shared_secret << endl;
	
	cout << (alices_shared_secret == bobs_shared_secret ? "match" : "doh") << endl;
}


int main(int argc, char **argv)
{
	//this is our "randomness" sources
	//(as it is not seeded, it will always generate the same sequence)
	gmp_randstate_t state;
	gmp_randinit_default(state);

	dh_example(state);
	
	dlp::DHSecretKey sk;
	dlp::DHPublicKey pk;
	
	sk.p = "345398920761083754448930116972501508061633731275362290546006767679820739731567";
	sk.g = 4;
	sk.o = "172699460380541877224465058486250754030816865637681145273003383839910369865783";
	sk.x = "126480516309396574217862884796244269755959298822681451445677342549367658410137";
	
	pk.set(sk);
	pk.y = "257196250433213704992737143678227617930551462806171346019162743609730142686792";
	
	ElgamalCiphertext c;
	
	c.kem_data = "55734020954639145669559113407539705356056779363148370512084530602269900039102";
	c.dem_data = "130740825623686848596680236139076340257614025191639446347975105246948429921846";
	
	mpz_class msg;
	
	c.dec(msg,sk);
	
	cout << "The password is:\n" << msg << endl;
    
    gmp_randclear(state);

	return 0;
}
